import logging

from requests import Response, Session, PreparedRequest

from .errors.access_error import AccessError
from .errors.auth_error import AuthError
from .errors.bad_request_error import BadRequestError
from .errors.base_http_error import BaseApiError
from .errors.not_found_error import NotFoundError
from .errors.server_error import ServerError
from .errors.unprocessable_entity_error import UnprocessableEntityError


class BaseHTTP:
    def __init__(self, session: Session, base_url, allure=None):
        self.session = session
        self.base_url = base_url
        self.allure = allure
        self.log = logging.getLogger(__name__)

    def get(self, url, **kwargs):
        callback = self.session.get
        return self._send(url, callback, 'GET', **kwargs)

    def post(self, url, **kwargs):
        callback = self.session.post
        return self._send(url, callback, 'POST', **kwargs)

    def put(self, url, **kwargs):
        callback = self.session.put
        return self._send(url, callback, 'PUT', **kwargs)

    def patch(self, url, **kwargs):
        callback = self.session.patch
        return self._send(url, callback, 'PATCH', **kwargs)

    def delete(self, url, **kwargs):
        callback = self.session.delete
        return self._send(url, callback, 'DELETE', **kwargs)

    def _send(self, url, callback, method, **kwargs):
        skip_body = kwargs.pop('skip_body', False)  # Need to pop this parameter before callback
        self._log_request(url, method, **kwargs)
        response = callback(self._path(url), **kwargs)
        self._log_response(response, skip_body)
        try:
            curl = to_curl(response.request, skip_headers=True)
            self._attach_to_allure(curl, 'curl')
            self.log.info(f'curl: {curl}')  # noqa: E800
        except UnicodeDecodeError:  # curlify can fall when decoding request body
            self.log.info('Cannot curlify request :(')

        self._raise_for_status(response)
        try:
            return response.json()
        except ValueError:
            return response.text

    def _log_request(self, url, method, **kwargs):
        request_message = f'{method} {self._path(url)}'
        if method == 'GET' and 'params' in kwargs:
            request_message += f'?{kwargs["params"]}'
        headers = {**self.session.headers, **kwargs.get('headers')} if kwargs.get('headers') else self.session.headers
        request_message += f'\nHeaders: {headers}'
        request_message += f'\nCookies: {self.session.cookies.get_dict()}'
        if kwargs.get('json'):
            request_message += f'\nBody: {kwargs.get("json")}'
        if kwargs.get('data'):
            request_message += f'\nData: {kwargs.get("data")}'
        if kwargs.get('files'):
            if isinstance(kwargs.get('files'), dict):  # single file uploading
                request_message += f'\nFiles: {kwargs.get("files")["file"][0]}'
            elif isinstance(kwargs.get('files'), list):  # multiple files uploading
                request_message += f'\nFiles: {[f[1][0] for f in kwargs.get("files")]}'
        self._attach_to_allure(request_message.replace('\n', '\n\n'), 'Request')
        self.log.info(request_message)

    def _log_response(self, response: Response, skip_body: bool):
        response_message = f'Response status: {response.status_code}, elapsed: {response.elapsed.seconds}s'
        response_message += f'\nResponse headers: {response.headers}'
        response_body = ''
        if skip_body:
            response_body = 'Response body was skipped'
        elif response.text != '':
            response_body = f'Response body: {self._decode(response.text)}'
        if response_body:
            response_message += f'\n{response_body}'
        self._attach_to_allure(response_message, 'Response')
        self.log.info(response_message)

    def _path(self, url):
        return f'{self.base_url}{url}'

    def _attach_to_allure(self, text, label):
        if self.allure:
            self.allure.attach(text, label, self.allure.attachment_type.TEXT)

    def _decode(self, text):
        return (text.encode('utf-8')).decode('utf-8')

    def _raise_for_status(self, response):
        if 400 <= response.status_code < 500:
            if response.status_code == 400:
                raise BadRequestError(response)
            elif response.status_code == 401:
                raise AuthError(response)
            elif response.status_code == 403:
                raise AccessError(response)
            elif response.status_code in [404, 410]:
                raise NotFoundError(response)
            elif response.status_code == 422:
                raise UnprocessableEntityError(response)
            else:
                raise BaseApiError('Client', response)
        elif 500 <= response.status_code < 600:
            raise ServerError(response)


def to_curl(request: PreparedRequest, compressed: bool = False, verify: bool = True, skip_headers: bool = True) -> str:
    """ Inner implementation of curlify until new version with https://github.com/ofw/curlify/pull/27

    Parameters
    ----------
        :param request: PreparedRequest object from requests.Response
        :param compressed: If `True` then `--compressed` argument will be added to result
        :param verify: If `True` then `--insecure` argument will be added to result
        :param skip_headers: If 'True' then headers [Accept, Accept-Encoding, Connection, User-Agent] will be skipped
    """
    curl = f'curl -X {request.method}'

    sys_headers = ['accept', 'accept-encoding', 'connection', 'user-agent', 'content-length']
    for k, v in request.headers.items():
        if skip_headers and k.lower() in sys_headers:
            continue
        curl += f" -H '{k}: {v}'"

    if request.body:
        body = request.body
        if isinstance(body, bytes):
            body = body.decode('utf-8')
        curl += f" -d '{body}'"

    curl += f' {request.url}'

    if compressed:
        curl += ' --compressed'
    if not verify:
        curl += ' --insecure'

    return curl



from .access_error import AccessError
from .auth_error import AuthError
from .bad_request_error import BadRequestError
from .base_http_error import BaseApiError
from .not_found_error import NotFoundError
from .server_error import ServerError
from .unprocessable_entity_error import UnprocessableEntityError



import logging

from google.protobuf.json_format import MessageToDict
from grpc import (RpcError,
                  StreamUnaryClientInterceptor, StreamStreamClientInterceptor,
                  UnaryUnaryClientInterceptor, UnaryStreamClientInterceptor)

from .grpc_error import GRPCError


class BaseInterceptor(UnaryUnaryClientInterceptor, UnaryStreamClientInterceptor,
                      StreamUnaryClientInterceptor, StreamStreamClientInterceptor):
    def __init__(self, channel, allure):
        self.allure = allure
        self.logger = logging.getLogger(__name__)
        self.target = channel._channel.target().decode()  # didn't find other way to get host

    def _intercept(self, continuation, client_call_details, request_or_iterator):
        self._log_request(self.target, client_call_details.method, request_or_iterator)

        response = continuation(client_call_details, request_or_iterator)

        self._raise_for_status(response)
        self._log_response(response)

        return response

    def _log_request(self, target, method, data):
        body = MessageToDict(data)
        self.logger.info(f'gRPC {target} {method} {body}')
        if self.allure:
            self.allure.attach(f'{target} {method} {body}', 'Request', self.allure.attachment_type.TEXT)

    def _log_response(self, response):
        body = MessageToDict(response.result())
        self.logger.info(f'Response: {body}')
        if self.allure:
            self.allure.attach(f'{body}', 'Response', self.allure.attachment_type.TEXT)

    def _raise_for_status(self, response):
        if not isinstance(response, RpcError):
            return
        self.logger.info(f'Response status: {response.code().name}, message: `{response.details()}`')
        if self.allure:
            self.allure.attach(response.details(), 'Error', self.allure.attachment_type.TEXT)
        raise GRPCError(status_code=response.code(), details=response.details())

    def intercept_unary_unary(self, continuation, client_call_details, request):
        return self._intercept(continuation, client_call_details, request)

    def intercept_unary_stream(self, continuation, client_call_details, request):
        return self._intercept(continuation, client_call_details, request)

    def intercept_stream_unary(self, continuation, client_call_details, request_iterator):
        return self._intercept(continuation, client_call_details, request_iterator)

    def intercept_stream_stream(self, continuation, client_call_details, request_iterator):
        return self._intercept(continuation, client_call_details, request_iterator)



from requests import HTTPError


class BaseApiError(HTTPError):
    def __init__(self, label, response):
        msg = f'{response.status_code} {label} Error for {response.url}'
        if response.text != '':
            msg = f'{msg}: {response.text}'
        super().__init__(msg, response=response)

    def body(self):
        try:
            return self.response.json()
        except ValueError:
            return self.response.text

    def reason(self):
        if isinstance(self.response.reason, bytes):
            try:
                reason = self.response.reason.decode('utf-8')
            except UnicodeDecodeError:
                reason = self.response.reason.decode('iso-8859-1')
        else:
            reason = self.response.reason
        return reason



from google.protobuf.json_format import MessageToDict, MessageToJson
from grpc import intercept_channel

from .base_interceptor import BaseInterceptor


class BaseStub:
    def __init__(self, channel, stub, allure=None):
        self.channel = intercept_channel(channel, BaseInterceptor(channel, allure))
        self.stub = stub(self.channel)


def response_to_dict(func):
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        return MessageToDict(response)

    return wrapper


def response_to_json(func):
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        return MessageToJson(response)

    return wrapper



from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry


class RetrySession(Session):
    def __init__(self, retries=3, backoff_factor=0.5, status_forcelist=None):
        super().__init__()
        retry = Retry(total=retries, read=retries, backoff_factor=backoff_factor, status_forcelist=status_forcelist)
        adapter = HTTPAdapter(max_retries=retries)
        self.mount('http://', adapter)
        self.mount('https://', adapter)
